@@ -2,9 +2,9 @@
 
 from contextlib import ExitStack
 from functools import partial
-import math
 import numpy as np
 import warnings
+from numbers import Number
 
 from affine import Affine
 
@@ -78,7 +78,7 @@ class TransformMethodsMixin:
         if transform_method is TransformMethod.gcps:
             transform = transform[0]
         if not transform:
-            raise AttributeError("Dataset has no {}".format(transform_method))
+            raise AttributeError(f"Dataset has no {transform_method}")
         return xy(transform, row, col, zs=z, offset=offset, **rpc_options)
 
     def index(
@@ -86,7 +86,7 @@ class TransformMethodsMixin:
         x,
         y,
         z=None,
-        op=math.floor,
+        op=np.floor,
         precision=None,
         transform_method=TransformMethod.affine,
         **rpc_options
@@ -130,8 +130,8 @@ class TransformMethodsMixin:
         if transform_method is TransformMethod.gcps:
             transform = transform[0]
         if not transform:
-            raise AttributeError("Dataset has no {}".format(transform_method))
-        return rowcol(transform, x, y, zs=z, op=op, **rpc_options)
+            raise AttributeError(f"Dataset has no {transform_method}")
+        return tuple(int(val) for val in rowcol(transform, x, y, zs=z, op=op, **rpc_options))
 
 
 def get_transformer(transform, **rpc_options):
@@ -251,7 +251,7 @@ def xy(transform, rows, cols, zs=None, offset='center', **rpc_options):
         return transformer.xy(rows, cols, zs=zs, offset=offset)
 
 
-def rowcol(transform, xs, ys, zs=None, op=math.floor, precision=None, **rpc_options):
+def rowcol(transform, xs, ys, zs=None, op=np.floor, precision=None, **rpc_options):
     """Get rows and cols of the pixels containing (x, y).
 
     Parameters
@@ -330,13 +330,19 @@ class TransformerBase:
         TransformError
             If input coordinates are not all of the same length
         """
+        xs = np.atleast_1d(xs)
+        ys = np.atleast_1d(ys)
+        if zs is not None:
+            zs = np.atleast_1d(zs)
+        else:
+            zs = np.zeros(1)
+
         try:
-            xs, ys, zs = np.broadcast_arrays(xs, ys, 0 if zs is None else zs)
+            broadcasted = np.broadcast(xs, ys, zs)
         except ValueError as error:
-            raise TransformError(
-                "Input coordinates must be broadcastable to a 1d array"
-            ) from error
-        return np.atleast_1d(xs), np.atleast_1d(ys), np.atleast_1d(zs)
+            raise TransformError() from error
+
+        return xs, ys, zs
 
     def __enter__(self):
         return self
@@ -344,7 +350,7 @@ class TransformerBase:
     def __exit__(self, *args):
         pass
 
-    def rowcol(self, xs, ys, zs=None, op=math.floor, precision=None):
+    def rowcol(self, xs, ys, zs=None, op=np.floor, precision=None):
         """Get rows and cols coordinates given geographic coordinates.
 
         Parameters
@@ -378,7 +384,7 @@ class TransformerBase:
                 RasterioDeprecationWarning,
             )
 
-        AS_ARR = True if hasattr(xs, "__iter__") else False
+        IS_SCALAR = isinstance(xs, Number) and isinstance(ys, Number)
         xs, ys, zs = self._ensure_arr_input(xs, ys, zs=zs)
 
         try:
@@ -386,10 +392,18 @@ class TransformerBase:
                 xs, ys, zs, transform_direction=TransformDirection.reverse
             )
 
-            if len(new_rows) == 1 and not AS_ARR:
-                return (op(new_rows[0]), op(new_cols[0]))
+            is_op_ufunc = isinstance(op, np.ufunc)
+            if is_op_ufunc:
+                op(new_rows, out=new_rows)
+                op(new_cols, out=new_cols)
             else:
-                return ([op(r) for r in new_rows], [op(c) for c in new_cols])
+                new_rows = np.array(list(map(op, new_rows)))
+                new_cols = np.array(list(map(op, new_cols)))
+
+            if IS_SCALAR:
+                return new_rows[0], new_cols[0]
+            else:
+                return new_rows, new_cols
         except TypeError:
             raise TransformError("Invalid inputs")
 
@@ -407,7 +421,8 @@ class TransformerBase:
             transformations. Default: 0.
         offset : str, optional
             Determines if the returned coordinates are for the center of the
-            pixel or for a corner.
+            pixel or for a corner. Available options include center, ul, ur, ll,
+            lr.
         Raises
         ------
         ValueError
@@ -418,7 +433,7 @@ class TransformerBase:
         tuple of float or list of float
 
         """
-        AS_ARR = True if hasattr(rows, "__iter__") else False
+        IS_SCALAR = isinstance(rows, Number) and isinstance(cols, Number)
         rows, cols, zs = self._ensure_arr_input(rows, cols, zs=zs)
 
         if offset == 'center':
@@ -434,24 +449,21 @@ class TransformerBase:
         else:
             raise TransformError("Invalid offset")
 
-        # shift input coordinates according to offset
-        T = IDENTITY.translation(coff, roff)
-        offset_rows = []
-        offset_cols = []
         try:
-            for colrow in zip(cols, rows):
-                offset_col, offset_row = T * colrow
-                offset_rows.append(offset_row)
-                offset_cols.append(offset_col)
-
+            # shift input coordinates according to offset
+            T = IDENTITY.translation(coff, roff)
+            identity_transformer = AffineTransformer(T)
+            offset_cols, offset_rows = identity_transformer._transform(
+                cols, rows, zs, transform_direction=TransformDirection.forward
+            )
             new_xs, new_ys = self._transform(
                 offset_cols, offset_rows, zs, transform_direction=TransformDirection.forward
             )
 
-            if len(new_xs) == 1 and not AS_ARR:
-                return (new_xs[0], new_ys[0])
+            if IS_SCALAR:
+                return new_xs[0], new_ys[0]
             else:
-                return (new_xs, new_ys)
+                return new_xs, new_ys
         except TypeError:
             raise TransformError("Invalid inputs")
 
@@ -483,22 +495,23 @@ class AffineTransformer(TransformerBase):
         if not isinstance(affine_transform, Affine):
             raise ValueError("Not an affine transform")
         self._transformer = affine_transform
+        self._transform_arr = np.empty((3, 3), dtype='float64')
 
     def _transform(self, xs, ys, zs, transform_direction):
-        resxs = []
-        resys = []
-
+        transform = self._transform_arr
         if transform_direction is TransformDirection.forward:
-            transform = self._transformer
+            transform.flat[:] = self._transformer
         elif transform_direction is TransformDirection.reverse:
-            transform = ~self._transformer
+            transform.flat[:] = ~self._transformer
 
-        for x, y in zip(xs, ys):
-            resx, resy = transform * (x, y)
-            resxs.append(resx)
-            resys.append(resy)
+        bi = np.broadcast(xs, ys)
+        input_matrix = np.empty((3, bi.size))
+        input_matrix[0] = bi.iters[0]
+        input_matrix[1] = bi.iters[1]
+        input_matrix[2] = 1
+        transform.dot(input_matrix, out=input_matrix)
+        return input_matrix[0], input_matrix[1]
 
-        return (resxs, resys)
 
     def __repr__(self):
         return "<AffineTransformer>"
@@ -532,13 +545,14 @@ class GCPTransformer(GCPTransformerBase, GDALTransformerBase):
 
     Uses GDALCreateGCPTransformer and GDALGCPTransform for computations.
     Ensure that GDAL transformer objects are destroyed by calling `close()`
-    method or using context manager interface.
+    method or using context manager interface. If `tps` is set to True,
+    uses GDALCreateTPSTransformer and GDALTPSTransform instead.
 
     """
-    def __init__(self, gcps):
+    def __init__(self, gcps, tps=False):
         if len(gcps) and not isinstance(gcps[0], GroundControlPoint):
             raise ValueError("GCPTransformer requires sequence of GroundControlPoint")
-        super().__init__(gcps)
+        super().__init__(gcps, tps)
 
     def __repr__(self):
         return "<{} GCPTransformer>".format(
