@@ -528,8 +528,7 @@ def validate_length_value(instance, attribute, value):
         raise ValueError("Number of columns or rows must be non-negative")
 
 
-@attr.s(slots=True,
-        frozen=True)
+@attr.s(slots=True, frozen=True)
 class Window:
     """Windows are rectangular subsets of rasters.
 
@@ -709,8 +708,8 @@ class Window:
     def round_lengths(self, **kwds):
         """Return a copy with width and height rounded.
 
-        Lengths are rounded to the preceding (floor) or succeeding (ceil)
-        whole number. The offsets are not changed.
+        Lengths are rounded to the nearest whole number. The offsets are
+        not changed.
 
         Parameters
         ----------
@@ -737,8 +736,8 @@ class Window:
     def round_offsets(self, **kwds):
         """Return a copy with column and row offsets rounded.
 
-        Offsets are rounded to the preceding (floor) or succeeding (ceil)
-        whole number. The lengths are not changed.
+        Offsets are rounded to the preceding whole number. The lengths
+        are not changed.
 
         Parameters
         ----------
@@ -755,6 +754,19 @@ class Window:
         col_off = operator(self.col_off)
         return Window(col_off, row_off, self.width, self.height)
 
+    def round(self, ndigits=None):
+        """Round a window's offsets and lengths
+
+        Rounding to a very small fraction of a pixel can help treat
+        floating point issues arising from computation of windows.
+        """
+        return Window(
+            round(self.col_off, ndigits=ndigits),
+            round(self.row_off, ndigits=ndigits),
+            round(self.width, ndigits=ndigits),
+            round(self.height, ndigits=ndigits),
+        )
+
     def crop(self, height, width):
         """Return a copy cropped to height and width"""
         return crop(self, height, width)
@@ -773,3 +785,50 @@ class Window:
         Window
         """
         return intersection([self, other])
+
+
+def subdivide(window, height, width):
+    """Divide a window into smaller windows.
+
+    Windows have no overlap and will be at most the desired
+    height and width. Smaller windows will be generated where
+    the height and width do not evenly divide the window dimensions.
+
+    Parameters
+    ----------
+    window : Window
+        Source window to subdivide.
+    height : int
+        Subwindow height.
+    width : int
+        Subwindow width.
+
+    Returns
+    -------
+    list of Windows
+    """
+    subwindows = []
+
+    irow = window.row_off + window.height
+    icol = window.col_off + window.width
+
+    row_off = window.row_off
+    col_off = window.col_off
+    while row_off < irow:
+        if row_off + height > irow:
+            _height = irow - row_off
+        else:
+            _height = height
+
+        while col_off < icol:
+            if col_off + width > icol:
+                _width = icol - col_off
+            else:
+                _width = width
+
+            subwindows.append(Window(col_off, row_off, _width, _height))
+            col_off += width
+
+        row_off += height
+        col_off = window.col_off
+    return subwindows
\ No newline at end of file
