@@ -5,7 +5,8 @@ import numpy as np
 import pytest
 
 import rasterio
-from rasterio.errors import DatasetIOShapeError
+from rasterio._err import CPLE_AppDefinedError
+from rasterio.errors import DatasetIOShapeError, RasterioIOError
 
 # Find out if we've got HDF support (needed below).
 try:
@@ -333,3 +334,32 @@ def test_read_out_mask(path_rgb_byte_tif, out):
     with rasterio.open(path_rgb_byte_tif) as src:
         with pytest.raises(ValueError):
             src.read(indexes=[2], out=out)
+
+
+def test_chained_io_errors(path_rgb_byte_tif):
+    """Get chained exceptions."""
+    with rasterio.open("tests/data/corrupt.tif") as src:
+        # RasterioIOError is at the top of the stack (~0).
+        with pytest.raises(RasterioIOError) as excinfo:
+            src.read()
+
+        assert "Read failed. See previous exception for details." == str(excinfo.value)
+
+        # Exception ~1 is a GDAL AppDefinedError mentioning IReadBlock.
+        exc = excinfo.value.__cause__
+        assert isinstance(exc, CPLE_AppDefinedError)
+        msg = str(exc)
+        assert "corrupt.tif" in msg
+        assert "IReadBlock failed" in msg
+
+        # Exception ~2 is another AppDefinedError mentioning TIFFReadEncodedTile.
+        exc = excinfo.value.__cause__.__cause__
+        assert isinstance(exc, CPLE_AppDefinedError)
+        msg = str(exc)
+        assert "TIFFReadEncodedTile()" in msg
+
+        # Exception ~3 is another AppDefinedError mentioning TIFFFillTile.
+        exc = excinfo.value.__cause__.__cause__.__cause__
+        assert isinstance(exc, CPLE_AppDefinedError)
+        msg = str(exc)
+        assert "TIFFFillTile:Read error" in msg
