@@ -14,7 +14,7 @@ import pytest
 
 import rasterio
 from rasterio.enums import Resampling
-from rasterio.merge import merge
+from rasterio.merge import merge, MERGE_METHODS
 from rasterio.rio.main import main_group
 from rasterio.transform import Affine
 
@@ -152,9 +152,9 @@ def test_merge_with_nodata(test_data_dir_1, runner):
 
 
 @pytest.mark.filterwarnings("ignore:Input file's nodata value")
-def test_merge_error(test_data_dir_1, runner):
+def test_merge_error(test_data_dir_1, tmp_path, runner):
     """A nodata value outside the valid range results in an error"""
-    outputname = str(test_data_dir_1.join('merged.tif'))
+    outputname = tmp_path.joinpath('merged.tif').as_posix()
     inputs = [str(x) for x in test_data_dir_1.listdir()]
     inputs.sort()
     with pytest.warns(UserWarning):
@@ -420,6 +420,7 @@ def test_merge_tiny_base(tiffs, runner):
 
     with rasterio.open(outputname) as src:
         data = src.read()
+        print(data)
         assert (data[0][0:2, 1] == 120).all()
         assert (data[0][0:2, 2:4] == 90).all()
         assert data[0][2][1] == 60
@@ -484,24 +485,46 @@ def test_merge_out_of_range_nodata(tiffs):
     assert datasets[1].dtypes[0] == 'uint8'
 
     with pytest.warns(UserWarning):
-        rv, transform = merge(datasets, nodata=9999)
+        rv, transform = merge(datasets, nodata=9999, masked=True)
+
+    assert rv.fill_value == 0.0
 
 
-def test_merge_rgb(tmpdir, runner):
+@pytest.mark.parametrize("res_option", [[], ["--use-first-res"], ["--use-highest-res"]])
+def test_merge_rgb(tmpdir, runner, res_option):
     """Get back original image"""
     outputname = str(tmpdir.join('merged.tif'))
     inputs = [
-        'tests/data/rgb1.tif',
-        'tests/data/rgb2.tif',
-        'tests/data/rgb3.tif',
-        'tests/data/rgb4.tif']
-    result = runner.invoke(main_group, ['merge'] + inputs + [outputname])
+        "tests/data/rgb1.tif",
+        "tests/data/rgb2.tif",
+        "tests/data/rgb3.tif",
+        "tests/data/rgb4.tif",
+    ]
+    result = runner.invoke(main_group, ["merge"] + res_option + inputs + [outputname])
     assert result.exit_code == 0
 
     with rasterio.open(outputname) as src:
         assert [src.checksum(i) for i in src.indexes] == [25420, 29131, 37860]
 
 
+@pytest.mark.parametrize("res_option", [["--use-highest-res"]])
+def test_merge_highest_res(tmpdir, runner, res_option):
+    """Ignore the resolution of the first source."""
+    outputname = str(tmpdir.join('merged.tif'))
+    inputs = [
+        "tests/data/rgb-byte-tenth.tif",
+        "tests/data/rgb1.tif",
+        "tests/data/rgb2.tif",
+        "tests/data/rgb3.tif",
+        "tests/data/rgb4.tif",
+    ]
+    result = runner.invoke(main_group, ["merge"] + res_option + inputs + [outputname])
+    assert result.exit_code == 0
+
+    with rasterio.open(outputname) as src:
+        assert tuple(round(x) for x in src.res) == (300.0, 300.0)
+
+
 def test_merge_tiny_intres(tiffs):
     inputs = [str(x) for x in tiffs.listdir()]
     inputs.sort()
@@ -565,7 +588,7 @@ def test_merge_precision(tmpdir, precision):
         # Compare header lines.
         for i in range(5):
             assert out_file.readline().strip() == expected_file.readline().strip()
-
+        
         # Compare raster data as single strings.
         out_data = " ".join(line.strip() for line in out_file.readlines())
         expected_data = " ".join(line.strip() for line in expected_file.readlines())
@@ -608,6 +631,7 @@ def test_merge_resampling(test_data_dir_resampling, resampling, runner):
         res = src.res[0]
         expected_raster = src.read(
             out_shape=tuple(dim * 2 for dim in src.shape),
+            boundless=True,
             resampling=resampling
         )
     result = runner.invoke(
@@ -713,3 +737,27 @@ def test_merge_no_gap(tiffs, runner):
     with rasterio.open(outputname) as src:
         data = src.read(1)
         assert data[184, 61] != 0
+
+
+@pytest.mark.parametrize(
+    "method",
+    list(MERGE_METHODS)
+)
+def test_rio_merge_method(test_data_dir_1, method, runner):
+    outputname = str(test_data_dir_1.join('merged.tif'))
+    inputs = [str(x) for x in test_data_dir_1.listdir()]
+
+    merged, _ = merge(
+        inputs, output_count=1, method=method, dtype=rasterio.uint8
+    )
+
+    result = runner.invoke(
+        main_group, ['merge'] + inputs + [outputname] +
+        ['--method', method])
+
+    assert result.exit_code == 0
+    assert os.path.exists(outputname)
+    with rasterio.open(outputname) as dst:
+        output_raster = dst.read()
+
+    np.testing.assert_array_equal(output_raster, merged)
