@@ -293,7 +293,7 @@ def test_xy_input(rows, cols, exp_xy, aff):
 
 
 @pytest.mark.parametrize("aff", [Affine.identity()])
-@pytest.mark.parametrize("rows, cols", [([0, 1, 2], [0, 1]), ("0", "0")])
+@pytest.mark.parametrize("rows, cols", [([0, 1, 2], [0, 1])])
 def test_invalid_xy_input(rows, cols, aff):
     """Raise on invalid input."""
     with pytest.raises(TransformError):
@@ -387,13 +387,14 @@ def test_rowcol_gcps_rpcs(dataset, transform_attr, coords, expected):
 def test_xy_rowcol_inverse(transform):
     # TODO this is an ideal candiate for
     # property-based testing with hypothesis
-    rows_cols = ([0, 0, 10, 10],
-                 [0, 10, 0, 10])
-    assert rows_cols == rowcol(transform, *xy(transform, *rows_cols))
+    rows = [0, 0, 10, 10]
+    cols = [0, 10, 0, 10]
+    crows, ccols = rowcol(transform, *xy(transform, rows, cols))
+    assert numpy.allclose(rows, crows) and numpy.allclose(cols, ccols)
 
 
 @pytest.mark.parametrize("aff", [Affine.identity()])
-@pytest.mark.parametrize("xs, ys", [([0, 1, 2], [0, 1]), ("0", "0")])
+@pytest.mark.parametrize("xs, ys", [([0, 1, 2], [0, 1])])
 def test_invalid_rowcol_input(xs, ys, aff):
     """Raise on invalid input."""
     with pytest.raises(TransformError):
@@ -436,7 +437,9 @@ def test_transformer_open_closed(transformer_cls, transform):
 )
 def test_ensure_arr_input(coords, expected):
     transformer = transform.AffineTransformer(Affine.identity())
-    assert transformer.xy(*coords, offset='ul') == expected
+    rows, cols = expected
+    crows, ccols = transformer.xy(*coords, offset='ul')
+    assert numpy.allclose(rows, crows) and numpy.allclose(cols, ccols)
 
 
 def test_ensure_arr_input_same_shape():
@@ -506,3 +509,24 @@ def test_2421_rpc_height_ignored():
         x2, y2 = src.xy(0, 0, z=2000, transform_method=transform_method)
         assert abs(x2 - x1) > 0
         assert abs(y2 - y1) > 0
+
+
+def test_gcp_transformer_tps_option():
+    """Use thin plate spline transformation when requested."""
+    # TPS ensures that GCPs are (to within some precision) solutions of the transformation.
+    # This is not the case for polynomials transformations. 
+    with GCPTransformer(gcps(), tps=True) as transformer:
+        for gcp in gcps():
+            x_, y_ = transformer.xy(gcp.row, gcp.col, offset='ul')
+            assert gcp.x == pytest.approx(x_)
+            assert gcp.y == pytest.approx(y_)
+            row_, col_ = transformer.rowcol(gcp.x, gcp.y, op=lambda arg: arg)
+            assert gcp.row == pytest.approx(row_)
+            assert gcp.col == pytest.approx(col_)
+
+
+def test_transform_grid():
+    """Accept a grid, see gh-3191."""
+    with rasterio.open('tests/data/RGB.byte.tif') as src:
+        cols, rows = numpy.mgrid[0:3, 0:3]
+        xs, ys = src.xy(cols, rows)
