@@ -7,7 +7,7 @@ import affine
 import numpy as np
 import pytest
 
-from .conftest import requires_gdal35, gdal_version
+from .conftest import gdal_version
 
 import rasterio
 from rasterio.drivers import blacklist
@@ -117,7 +117,7 @@ def test_write_sbyte(tmpdir):
 
     info = subprocess.check_output(["gdalinfo", "-stats", name]).decode('utf-8')
     assert "Minimum=-33.000, Maximum=-33.000, Mean=-33.000, StdDev=0.000" in info
-    if GDALVersion.runtime() < GDALVersion(3, 7):
+    if gdal_version < GDALVersion(3, 7):
         assert 'SIGNEDBYTE' in info
     else:
         assert 'Int8' in info
@@ -408,7 +408,7 @@ def test_too_big_for_tiff(tmpdir):
     ('jpeg', 'JPEG'),
 ])
 def test_write__autodetect_driver(tmpdir, extension, driver):
-    name = str(tmpdir.join("test.{}".format(extension)))
+    name = str(tmpdir.join(f"test.{extension}"))
     with rasterio.open(name, 'w', height=1, width=1, count=1, dtype='uint8') as rds:
         assert rds.driver == driver
 
@@ -533,7 +533,6 @@ def test_write_masked_nomask(tmp_path):
         assert list(arr.flatten()) == [0, 1, 2]
 
 
-@requires_gdal35
 def test_write_int64(tmp_path):
     test_file = tmp_path / "test.tif"
     data = np.array([np.ones((100, 100), dtype=rasterio.int64) * 127])
@@ -553,26 +552,6 @@ def test_write_int64(tmp_path):
         assert file.dtypes == (rasterio.int64,)
 
 
-@pytest.mark.skipif(
-    gdal_version.at_least('3.5'),
-    reason="Validate behavior before GDAL 3.5",
-)
-def test_write_int64__unsupported(tmp_path):
-    test_file = tmp_path / "test.tif"
-    data = np.array([np.ones((100, 100), dtype=rasterio.int64) * 127])
-    with pytest.raises(TypeError, match="invalid dtype"):
-        with rasterio.open(
-            test_file,
-            'w',
-            driver='GTiff',
-            width=100,
-            height=100,
-            count=1,
-            dtype=data.dtype
-        ) as file:
-            file.write(data, [1])
-
-
 def test_open_no_log(caplog, tmp_path):
     """See gh-2525."""
     caplog.set_level(logging.DEBUG)
